#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include "utf8.h"
#include "EasyString.h"

char* Private_Substring(char *String, size_t Position, size_t Length)
{
	char *OutputBuffer;
	size_t TotalSizeA = 0;
	size_t TotalSizeB = 0;
	size_t TotalLength = utf8len(String);

	if ((Position + Length) > TotalLength)
	{
		Length = TotalLength - Position;
	}

	for (int i = 0; i < Position; i++)
	{
		TotalSizeA += utf8codepointcalcsize(String + sizeof(char)*i);
	}
	
	char* Cursor = String + TotalSizeA;
	
	for (int i = 0; i < Length; i++)
	{
		TotalSizeB += utf8codepointcalcsize(Cursor + sizeof(char)*i);
	}
	
	OutputBuffer = calloc(TotalSizeB+1, sizeof(char));
   
	if( OutputBuffer == NULL )
	{
		exit(EXIT_FAILURE);
	}
	
	memcpy(OutputBuffer, Cursor, TotalSizeB);
	
	return OutputBuffer;
}

char* EasyString_GetBuffer(EasyString* This)
{
	return This->Buffer;
}

void EasyString_Clear(EasyString* This)
{
	
	if (This->Buffer != NULL && This->FreeBuffer == true) 
	{
		free(This->Buffer);
	}
	
	This->Buffer = NULL;
	This->Capacity = 0;
	This->FreeBuffer = false;
	
}

void EasyString_SetConst(EasyString* This, char* Buffer)
{
	EasyString_Clear(This);
	
    This->Buffer = Buffer;

    if (This->Buffer != NULL)
    {	
		This->Capacity = utf8size(This->Buffer);
	}
	
}

void EasyString_SetBuffer(EasyString* This, char* Buffer)
{
	EasyString_Clear(This);
    This->Buffer = Buffer;
    
    if (This->Buffer != NULL)
    {	
		This->Capacity = utf8size(This->Buffer);
		This->FreeBuffer = true;
	}
}

size_t EasyString_Length(EasyString* This)
{
	return utf8len(This->Buffer);
}

size_t EasyString_Capacity(EasyString* This)
{
	return This->Capacity;
}

void EasyString_Append(EasyString* This, char* Buffer)
{
	char* TemporaryBuffer = NULL;
	
	if (This->Capacity == 0)
	{
		This->Capacity = 1;
	}
	
	This->Capacity = This->Capacity + utf8len(Buffer);
	
	if (This->FreeBuffer == false)
	{
		TemporaryBuffer = (char*) calloc(This->Capacity, sizeof(char));
		
		if (TemporaryBuffer == NULL)
		{
			exit(EXIT_FAILURE);	
		}
		
		if (This->Buffer == NULL)
		{
			This->Buffer = TemporaryBuffer;
			strcpy(TemporaryBuffer, Buffer);
		}
	    else
	    {
			strcpy(TemporaryBuffer, This->Buffer);
			strcat(TemporaryBuffer, Buffer);
			This->Buffer = TemporaryBuffer;
		}
		
		This->FreeBuffer = true;
		
	}
	else
	{
		TemporaryBuffer = (char*) calloc(This->Capacity, sizeof(char));
		
		if (TemporaryBuffer == NULL)
		{
			exit(EXIT_FAILURE);	
		}
		else
		{
			strcpy(TemporaryBuffer, This->Buffer);
			free(This->Buffer);
			strcat(TemporaryBuffer, Buffer);
			This->Buffer = TemporaryBuffer;
		}
	}
}

void EasyString_AppendChar(EasyString* This, char to_be_added)
{
	char buffer[2];
	buffer[0] = to_be_added;
	buffer[1] = '\0';
	
	EasyString_Append(This, buffer);
} 

void EasyString_Add(EasyString* This, EasyString* Other)
{
	EasyString_Append(This, Other->Buffer);
} 

void EasyString_Insert(EasyString* This, EasyString* Other, int Position)
{
	char* TemporaryBuffer = NULL;
	
	if (This->Capacity - 1 >= Position)
	{
		This->Capacity = This->Capacity + utf8len(Other->Buffer);

		if (This->FreeBuffer == false)
		{

			char *f, *e;
			int length;
			length = strlen(This->Buffer);
		   
			f = Private_Substring(This->Buffer, 0, Position);      
			e = Private_Substring(This->Buffer, Position, length - Position + 1);
			
			TemporaryBuffer = calloc(This->Capacity, sizeof(char));
			
			strcat(TemporaryBuffer, f);
			free(f);
			strcat(TemporaryBuffer, Other->Buffer);
			strcat(TemporaryBuffer, e);
			free(e);
			
			This->Buffer = TemporaryBuffer;
			This->FreeBuffer =true;
		}
		else
		{
			char *f, *e;
			int length;
			length = strlen(This->Buffer);
		   
			f = Private_Substring(This->Buffer, 0, Position);      
			e = Private_Substring(This->Buffer, Position, length - Position + 1);
			This->Buffer = realloc(This->Buffer, This->Capacity * sizeof(char));
			
			memset(This->Buffer, '\0', This->Capacity);
			strcat(This->Buffer, f);
			free(f);
			strcat(This->Buffer, Other->Buffer);
			strcat(This->Buffer, e);
			free(e);
			
		}
		
	}
}

void EasyString_Remove(EasyString* This, EasyString* Other)
{

    char* TemporaryBuffer = NULL;
    
    if (This->FreeBuffer == false)
    {
		TemporaryBuffer = calloc(This->Capacity, sizeof(char));
		strcpy(TemporaryBuffer, This->Buffer);
		
		char *p, *q, *r;
		
		if (*(Other->Buffer) && (q = r = strstr(TemporaryBuffer, Other->Buffer)) != NULL) 
		{
			size_t len = strlen(Other->Buffer);
			
			while ((r = strstr(p = r + len, Other->Buffer)) != NULL) 
			{
				while (p < r)
					*q++ = *p++;
			}
			while ((*q++ = *p++) != '\0')
				continue;
		}
    
		This->FreeBuffer = true;
		This->Capacity = strlen(This->Buffer) + 1;
		This->Buffer = realloc(TemporaryBuffer, This->Capacity * sizeof(char));
    
	}
    else
    {
		char *p, *q, *r;
		
		if (*(Other->Buffer) && (q = r = strstr(This->Buffer, Other->Buffer)) != NULL) 
		{
			size_t len = strlen(Other->Buffer);
			
			while ((r = strstr(p = r + len, Other->Buffer)) != NULL) 
			{
				while (p < r)
					*q++ = *p++;
			}
			while ((*q++ = *p++) != '\0')
				continue;
		}
    
		This->Capacity = strlen(This->Buffer) + 1;
		This->Buffer = realloc(This->Buffer, This->Capacity);
	}
 
}

void EasyString_UCase(EasyString* This)
{
	if (This->FreeBuffer == false)
	{
		char* TemporaryBuffer = calloc(This->Capacity, sizeof(char));
		strcpy(TemporaryBuffer, This->Buffer);
		This->Buffer = TemporaryBuffer;
		This->FreeBuffer = true;
		
	}
	
	utf8upr(This->Buffer);
}

void EasyString_LCase(EasyString* This)
{
	if (This->FreeBuffer == false)
	{
		char* TemporaryBuffer = calloc(This->Capacity, sizeof(char));
		strcpy(TemporaryBuffer, This->Buffer);
		This->Buffer = TemporaryBuffer;
		This->FreeBuffer = true;
		
	}
	
	utf8lwr(This->Buffer);
	
}

int EasyString_Compare(EasyString* This, EasyString* Other)
{
	if (This->Buffer != NULL && Other->Buffer != NULL)
	{
		return strcmp(This->Buffer, Other->Buffer);
	}
	
	return -2;
}

EasyString* EasyString_Copy(EasyString* This)
{
	EasyString* Copy = EasyString_New(NULL);
	
	EasyString_Append(Copy, This->Buffer);

	return Copy;
}

EasyString* EasyString_Left(EasyString* This, size_t CharCount)
{
	EasyString* Left = EasyString_New(NULL);
	size_t Length = utf8len(This->Buffer);
	size_t TotalSize = 0;
	size_t MinimumChar = CharCount;
	
	if (Length < MinimumChar)
	{
		MinimumChar = Length;
	}
	
	for (int i = 0; i < MinimumChar; i++)
	{
		TotalSize += utf8codepointcalcsize(This->Buffer + sizeof(char)*i);
	}
	
	char* Buffer = calloc(TotalSize+1, sizeof(char));
	memcpy(Buffer, This->Buffer, TotalSize);

	EasyString_SetBuffer(Left, Buffer);

	return Left;
}

EasyString* EasyString_Right(EasyString* This, size_t CharCount)
{
	EasyString* Right = EasyString_New(NULL);
	
	size_t Length = utf8len(This->Buffer);
	size_t TotalSize = 0;
	
	char* Cursor = (Length < CharCount ? This->Buffer : This->Buffer + Length - CharCount);
	
	size_t MinimumChar = CharCount;
	
	if (Length < MinimumChar)
	{
		MinimumChar = Length;
	}
	
	for (int i = 0; i < MinimumChar; i++)
	{
		TotalSize += utf8codepointcalcsize(Cursor + sizeof(char)*i);
	}

	char* Buffer = calloc(TotalSize+1, sizeof(char));
	
	memcpy(Buffer, Cursor, TotalSize);

	EasyString_SetBuffer(Right, Buffer);

	return Right;
}

EasyString* EasyString_Mid(EasyString* This, size_t StartPosition, size_t Length)
{
	EasyString* Mid = EasyString_New(NULL);
	
	EasyString_SetBuffer(Mid, Private_Substring(This->Buffer, StartPosition, Length));
	
	return Mid;
}

EasyString* EasyString_Split(EasyString* This, size_t Index, EasyString* Delimiter)
{
	EasyString* Split = EasyString_New(NULL);
	
	char* Buffer = calloc(This->Capacity, sizeof(char));
	strcpy(Buffer, This->Buffer);
	
	char* Token = strtok(Buffer, Delimiter->Buffer);

	for (size_t i=1; i<Index; i++)
	{
		Token = strtok(NULL, Delimiter->Buffer);
	}
	
	size_t size = utf8size(Token);
	char* SplitBuffer = calloc(size, sizeof(char));
	strcpy(SplitBuffer, Token);
	
	EasyString_SetBuffer(Split, SplitBuffer);
	
	free(Buffer);
	
	return Split;
}

void EasyString_StrF(EasyString* This, float Value)
{
	int len = snprintf(NULL, 0, "%.8f", Value);
	char* Buffer = calloc(len+1, sizeof(char));
	
	snprintf(Buffer, len+1, "%.8f", Value);
	EasyString_SetBuffer(This,Buffer);
	
}

void EasyString_StrD(EasyString* This, double Value)
{
	int len = snprintf(NULL, 0, "%.9lf", Value);
	char* Buffer = calloc(len+1, sizeof(char));
	
	snprintf(Buffer, len+1, "%.9lf", Value);
	EasyString_SetBuffer(This,Buffer);
	
}

float EasyString_ValF(EasyString* This)
{
	char* End_Ptr;
	float Value = strtof(This->Buffer, &End_Ptr);

	return Value;
}

double EasyString_ValD(EasyString* This)
{
	char* End_Ptr;
	double Value = strtod(This->Buffer, &End_Ptr);
	
	return Value;
}

void EasyString_Move(EasyString* This, EasyString* Other)
{
	EasyString_Clear(This);
	EasyString_Append(This, Other->Buffer);
}

void EasyString_Destructor(void* to_be_destroyed)
{
	EasyString_Clear((EasyString*) to_be_destroyed);
	free(to_be_destroyed);
}

static const EasyStringVirtualTable easy_string_virtual_table = {
	.GetBuffer = EasyString_GetBuffer,
	.Clear = EasyString_Clear,
	.SetConst = EasyString_SetConst,
	.SetBuffer = EasyString_SetBuffer,
	.Length = EasyString_Length,
	.Capacity = EasyString_Capacity,
	.Append = EasyString_Append,
	.AppendChar = EasyString_AppendChar,
	.Add = EasyString_Add,
	.Insert = EasyString_Insert,
	.Remove = EasyString_Remove,
	.UCase = EasyString_UCase,
	.LCase = EasyString_LCase,
	.Compare = EasyString_Compare,
	.Copy = EasyString_Copy,
	.Left = EasyString_Left,
	.Right = EasyString_Right,
	.Mid = EasyString_Mid,
	.Split = EasyString_Split,
	.StrF = EasyString_StrF,
	.StrD = EasyString_StrD,
	.ValF = EasyString_ValF,
	.ValD = EasyString_ValD,
	.Move = EasyString_Move,
	.Wipeout = EasyString_Clear,
    .Destructor = EasyString_Destructor
};

void EasyString_Init(EasyString* This, char* Buffer)
{
	This->vtable = &easy_string_virtual_table;
	
	This->Buffer = Buffer;
	
	if (This->Buffer != NULL)
	{	
		This->Capacity = strlen(Buffer) + 1;
	}
	
	This->FreeBuffer = false;
}


EasyString* EasyString_New(char* Buffer)
{
	EasyString* This = calloc(1, sizeof(EasyString));
	
	if (This == NULL)
	{
		exit(EXIT_FAILURE);	
	}
	
	EasyString_Init(This, Buffer);
	
	return This;
}

// <<<<<<<<<<<<<<<<<<<<<<<
// <<<<< END OF FILE <<<<<
// <<<<<<<<<<<<<<<<<<<<<<<
