#ifndef EASY_STRING_H
#define EASY_STRING_H

#include <stdlib.h>
#include <stdbool.h>

typedef struct EasyStringVirtualTable EasyStringVirtualTable;

typedef struct EasyString
{
	char* Buffer;
	size_t Capacity;
	bool FreeBuffer;
	const EasyStringVirtualTable *vtable;
} EasyString;

struct EasyStringVirtualTable 
{
   char* (*GetBuffer)(EasyString*);
   void (*Clear)(EasyString*);
   void (*SetConst)(EasyString*, char*);
   void (*SetBuffer)(EasyString*, char*);
   size_t (*Length)(EasyString*);
   size_t (*Capacity)(EasyString*);
   void (*Append)(EasyString*, char*);
   void (*AppendChar)(EasyString*, char);
   void (*Add)(EasyString*, EasyString*);
   void (*Insert)(EasyString*, EasyString*, int);
   void (*Remove)(EasyString*, EasyString*);
   void (*UCase)(EasyString*);
   void (*LCase)(EasyString*);
   int (*Compare)(EasyString*, EasyString*);
   EasyString* (*Copy)(EasyString*);
   EasyString* (*Left)(EasyString*, size_t);
   EasyString* (*Right)(EasyString*, size_t);
   EasyString* (*Mid)(EasyString*, size_t, size_t);
   EasyString* (*Split)(EasyString*, size_t, EasyString*);
   void (*StrF)(EasyString*, float);
   void (*StrD)(EasyString*, double);
   float (*ValF)(EasyString*);
   double (*ValD)(EasyString*);
   void (*Move)(EasyString*, EasyString*);
   void (*Wipeout)(EasyString*);
   void (*Destructor)(void*);
};

void EasyString_Destructor(void*);
void EasyString_Init(EasyString*, char*);

EasyString* EasyString_New(char*);

#endif

// <<<<<<<<<<<<<<<<<<<<<<<
// <<<<< END OF FILE <<<<<
// <<<<<<<<<<<<<<<<<<<<<<<
